# ewnlex.py
# Library to create and query a lexicon derived from Eurowordnet
#
# Copyright (C) 2004  Erwin Marsi
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
# Informationn on how to contact the author can be found in the file CONTACT.


"""
Library to create and query a lexicon derived from Eurowordnet

Notes:
- works for Dutch Eurowordnet files; not tested for other languages.
"""

__author__ = 'Erwin Marsi <e.c.marsi@uvt.nl>'
__version__ = '$Id: ewnlex.py,v 1.2 2004/07/13 19:26:22 emarsi Exp $'


# TODO:
# - improve calculation speed of closures



from sys import stdout, stderr
from string import join
from cPickle import dump, load
from sets import Set
from ewnparse import parse_ewn_file


class Lexicon(object):
    """
    A lexicon is what the name suggests. The keys are words, the
    values a Lemma objects. It can be filled by reading Eurowordnet
    database files or loading a previously pickled lexicon. It has
    methods to pretty print entries, but the interesting methods are
    for testing and retrieving lexical relations between words
    (e.g. synonym, hyponymn, hyperonym, etc.).
    """

    def __init__(self, fn=None):
        """
        L.__init__([FN])

        Create a new lexicon L, optionally loading it from file FN.
        """
        if fn:
            self.load(fn)
        else:
            self.d = {} 

    def read_ewn_file(self, fn):
        """
        L.read_ewn_file(FN)

        Reads a Eurowordnet database file in text format (.ewn),
        and adds it to lexicon L. 
        """
        
        # iterate over all WORD_MEANING objects parsed from the ewn file
        for wm in parse_ewn_file(fn):
            # list of word,sense pair in synset
            wsl = []
            pos =  wm['PART_OF_SPEECH']
            
            for var in wm['VARIANTS']:
                word =  var['STRING']
                sense = var['SENSE']
                wsl.append((word,sense))
                self.add_entry(word, pos, sense)

            for source_word, source_sense in wsl:

                # add synonymy relations among all members of the synset
                for dest_word, dest_sense in wsl:
                    if (source_word, source_sense) != (dest_word, dest_sense):
                        self.d[source_word].add_relation(pos,
                                                         source_sense,
                                                         'has_synonym',
                                                         dest_word,
                                                         wm['PART_OF_SPEECH'],
                                                         dest_sense)
                        
                # add other internal relations (not always present)
                if wm.has_key('INTERNAL_LINKS'):
                    for rel in  wm['INTERNAL_LINKS']:
                        self.d[source_word].add_relation(pos,
                                                         source_sense,
                                                         name = rel['NAME'],
                                                         dest_word = rel['TARGET_CONCEPT']['LITERAL']['STRING'],
                                                         dest_pos = rel['TARGET_CONCEPT']['PART_OF_SPEECH'],
                                                         dest_sense =  rel['TARGET_CONCEPT']['LITERAL']['SENSE'])

    def dump(self, fn, protocol=2):
        """
        L.dump(FN, [P=2])

        Write pickle of content of lexicon L to file FN
        (using pickle protocol P)
        """
        dump(self.d, open(fn, 'w'), protocol)

    def load(self, fn):
        """
        L.load(FN)

        Read pickled content of lexicon L from fle FN
        """
        self.d = load(open(fn))
                
    def add_entry(self, word, pos, sense):
        """
        L.add_entry(W, P, S)

        Add an entry for word W with part-of-spech P and sense S to
        lexicon L.
        """
        try:
            # extend the lemma with this pos (if necessary) and
            # extend its senses to include this sense
            # (if not already sufficient)
            self.d[word].extend_senses(pos, sense)
        except KeyError:
            # create a new lemma for this word, with pos and
            # enough senses to accomodate this sense
            self.d[word] = Lemma(pos, sense)
            
    def has_relation(self, word, relation, other_word,
                     partsos='', senses='',
                   other_pos='', other_sense=''):
        """
        L.has_relation(W1, R, W2, [PS='', SS='', P='', S=''])
        -> True or False
        
        Test if word W1 has relation R to other word W2, where W1 is
        optionally restricted by multiple parts-of-speech PS and/or
        senses SS, and W2 is optionally restricted by a part-of-speech
        P and/or sense S. PS and SS must be comma-separated strings.
        """
        try:
            return self.d[word].has_relation(relation, other_word,
                                             partsos, senses,
                                             other_pos, other_sense)
        except KeyError:
            False
            
    def related(self, word, relation, partsos='', senses=''):
        """
        L.related(W, R, [PS='', SS='']) -> list of related (word,pos,sense) tuples

        Retrieve all words that have relation R to word W, where W is
        optionally restricted by multiple parts of speech PS and/or
        senses SS. PS and SS must be comma-separated strings. The
        related words are returned as a tuple together with their part
        of speech and sense. The returned list may be empty.        
        """
        try:
            return self.d[word].related(relation, partsos, senses)
        except KeyError:
            return []
            
    def has_transitive_relation(self, word, relation, other_word,
                                partsos='', senses='',
                                other_pos='', other_sense=''):
        """
        L.has_transitive_relation(W1, R, W2, [PS='', SS='', P='', S=''])
        -> True or False
        
        Test if word W2 is in the transtive closure of relation R from
        word W1, where W1 is optionally restricted by multiple
        parts-of-speech PS and/or senses SS, and W2 is optionally
        restricted by a part-of-speech P and/or sense S. PS and SS
        must be comma-separated strings.
        """
        # this is certainly not the fastest implementation,
        # but it seems to avoid the problem of cycles,
        # because self.closure uses sets
        if other_sense:
            other_sense = int(other_sense)
            
        for w, p, s in self.closure(word, relation, partsos, senses):
            if ( w == other_word and
                 (not other_pos or p == other_pos) and
                 (not other_sense or s == other_sense) ):
                return True
            
        return False
        
    def closure(self, word, relation, partsos='', senses=''): 
        """
        L.closure(W, R, [PS, SS]) ->
        set of hyponym (word,pos,sense) tuples

        Return the transitive closure of relation R of word W, where W
        is optionally restricted by multiple parts of speech PS and/or
        senses SS. PS and SS must be comma-separated strings. The
        related words are returned as a tuple together with their part
        of speech and sense. The returned list may be empty.

        Note that some relations (e.g. 'has_synonym') may contain
        cycles. Since cycles are not checked for, this results in an
        endless loop (caught by the interpreter after some time).
        """       

        direct = Set(self.related(word, relation, partsos, senses))
        indirect = Set()
        for word, pos, sense in direct:
            indirect.union_update(self.closure(word, relation,
                                               pos, str(sense)))
        return indirect | direct

    def has_synonym(self, word, other_word,
                   partsos='', senses='',
                   other_pos='', other_sense=''):
        """
        L.has_synonym(W1, W2, [PS, SS, P, S]) -> True or False

        Test if W1 is a synonym of W2, where W1 is optionally
        restricted by multiple parts of speech PS and/or senses SS,
        and W2 is optionally restricted by a part-of-speech P and/or
        sense S. PS and SS must be comma-separated strings.
        """
        return self.has_relation(word, 'has_synonym', other_word,
                                 partsos, senses,
                                 other_pos, other_sense )
    
    def synonyms(self, word, partsos='', senses=''):
        """
        L.synonyms(W, [PS, SS]) -> list of synonym (word,pos,sense) tuples

        Retrieve all words that are a synonym of word W, where W is
        optionally restricted by multiple parts of speech PS and/or
        senses SS. PS and SS must be comma-separated strings. The
        related words are returned as a tuple together with their part
        of speech and sense. The returned list may be empty.        
        """
        return self.related(word, 'has_synonym', partsos, senses)

    def has_hyponym(self, word, other_word,
                    partsos='', senses='',
                    other_pos='', other_sense=''):
        """
        L.has_hyponym(W1, W2, [PS, SS, P, S]) -> True or False

        Test if W1 is a hynonym of W2, where W1 is optionally
        restricted by multiple parts of speech PS and/or senses SS,
        and W2 is optionally restricted by a part-of-speech P and/or
        sense S. PS and SS must be comma-separated strings.
        """
        return self.has_relation(word, 'has_hyponym', other_word,
                               partsos, senses,
                               other_pos, other_sense )

    def hyponyms(self, word, partsos='', senses=''):
        """
        L.hyponyms(W, [PS, SS]) -> list of hyponym (word,pos,sense) tuples

        Retrieve all words that are a hyponym of word W, where W is
        optionally restricted by multiple parts of speech PS and/or
        senses SS. PS and SS must be comma-separated strings. The
        related words are returned as a tuple together with their part
        of speech and sense. The returned list may be empty.        
        """
        return self.related(word,'has_hyponym', partsos, senses)
    
    def has_transitive_hyponym(self, word, other_word,
                               partsos='', senses='',
                               other_pos='', other_sense=''):
        """
        L.has_transitive_hyponym_closure(W1, W2, [PS, SS, P, S]) -> True or False

        Test if W2 is in the transitive closure of hyponyms of W1,
        where W1 is optionally restricted by multiple parts of speech
        PS and/or senses SS, and W2 is optionally restricted by a
        part-of-speech P and/or sense S. PS and SS must be
        comma-separated strings.
        
        """
        return self.has_transitive_relation(word, 'has_hyponym',
                                            other_word,
                                            partsos, senses,
                                            other_pos, other_sense)
    
    def hyponyms_closure(self, word, partsos='', senses=''):
        """
        L.hyponyms_closure(W, [PS, SS]) ->
        list of hyponym (word,pos,sense) tuples

        Return the transitive closure of hyponyms of word W, where W is
        optionally restricted by multiple parts of speech PS and/or
        senses SS. PS and SS must be comma-separated strings. The
        related words are returned as a tuple together with their part
        of speech and sense. The returned list may be empty.        
        """
        return list(self.closure(word, 'has_hyponym', partsos, senses))
    
    def has_hyperonym(self, word, other_word,
                      partsos='', senses='',
                      other_pos='', other_sense=''):
        """
        L.has_hyperonym(W1, W2, [PS, SS, P, S]) -> True or False

        Test if W1 is a hyperonym of W2, where W1 is optionally
        restricted by multiple parts of speech PS and/or senses SS,
        and W2 is optionally restricted by a part-of-speech P and/or
        sense S.  PS and SS must be comma-separated strings.
        """
        return self.has_relation(word, 'has_hyperonym', other_word,
                               partsos, senses,
                               other_pos, other_sense )
        
    def hyperonyms(self, word, partsos='', senses=''):
        """
        L.hyperonyms(W, [PS, SS]) -> list of hyperoym (word,pos,sense) tuples

        Retrieve all words that are a hyperonym of word W, where W is
        optionally restricted by multiple parts of speech PS and/or
        senses SS. PS and SS must be comma-separated strings. The
        related words are returned as a tuple together with their part
        of speech and sense. The returned list may be empty.        
        """
        return self.related(word, 'has_hyperonym', partsos, senses)
    
    def has_transitive_hyperonym(self, word, other_word,
                                 partsos='', senses='',
                                 other_pos='', other_sense=''):
        """
        L.has_transitive_hyperonym_closure(W1, W2, [PS, SS, P, S]) ->
        True or False

        Test if W2 is in the transitive closure of hyperonyms of W1,
        where W1 is optionally restricted by multiple parts of speech
        PS and/or senses SS, and W2 is optionally restricted by a
        part-of-speech P and/or sense S. PS and SS must be
        comma-separated strings.
        
        """
        return self.has_transitive_relation(word, 'has_hyperonym',
                                            other_word,
                                            partsos, senses,
                                            other_pos, other_sense)
    
    def hyperonyms_closure(self, word, partsos='', senses=''):
        """
        L.hyperonyms_closure(W, [PS, SS]) ->
        list of hyperonym (word,pos,sense) tuples

        Return the transitive closure of hyperonyms of word W, where W
        is optionally restricted by multiple parts of speech PS and/or
        senses SS. PS and SS must be comma-separated strings. The
        related words are returned as a tuple together with their part
        of speech and sense. The returned list may be empty.
        """
        return list(self.closure(word, 'has_hyperonym', partsos, senses))

    def write(self, s=stdout):
        """
        L.write([S=stdout])

        Write all entries in lexicon L to stream S 
        """
        words = self.d.keys()
        words.sort()
        for w in words:
            self.write_entry(w, s)
            
    def write_entry(self, word, s=stdout):
        """
        L.write_entry([W, S=stdout])

        Write entry for word W in lexicon L to stream S 
        """
        s.write('Word: %s\n' % word)
        self.d[word].write(s, 2)

    
        
class Lemma(dict):
    """
    A lemma object represents the properties of a word in the
    lexicon.

    A word can occur with one or more part-of-speech (pos)
    labels. These are the keys of the lemma dict.

    Any combination word + pos can have multiple senses. These are the
    lists that form the values of the lemma dict.

    Each word + pos + sense combination can have a number of relations
    to other word + pos + sense combination. The relation names are
    the values of the dicts in the sense lists. The values of these
    dicts are list (word,pos,sense) tuples.

    Take a look at the lemma information for the word 'programmeren'
    ('to program'):
    
    Word: programmeren
      Part-of-speech: v
        Sense: 1
          Relation: has_hyponym:
            voorprogrammeren (v,1)
          Relation: has_hyperonym:
            ontwerpen (v,1)
        Sense: 2
          Relation: involved_result:
            planning (n,1)
          Relation: has_hyperonym:
            ontwerpen (v,1)
            regelen (v,1)
          Relation: has_hyponym:
            afspreken (v,1)
            omprogrammeren (v,1)
            timen (v,2)
          Relation: has_synonym:
            plannen (v,1)
            uitstippelen (v,1)
          Relation: xpos_near_synonym:
            planning (n,2)
          Relation: is_subevent_of:
            leiden (v,4)
            beleid (n,1)

    It's Lemma object has the following structure:

    {'v':
      [
        {'has_hyponym':
          [
            ('voorprogrammeren', 'v', 1)],
         'has_hyperonym':
           [
             ('ontwerpen', 'v', 1)]},
        {'involved_result':
           [
             ('planning', 'n', 1)],
         'has_hyperonym':
           [
             ('ontwerpen', 'v', 1),
             ('regelen', 'v', 1)],
         'has_hyponym':
           [
             ('afspreken', 'v', 1),
             ('omprogrammeren', 'v', 1),
             ('timen', 'v', 2)],
         'has_synonym':
           [
             ('plannen', 'v', 1),
             ('uitstippelen', 'v', 1)],
         'xpos_near_synonym':
           [
             ('planning', 'n', 2)],
         'is_subevent_of':
           [
             ('leiden', 'v', 4),
             ('beleid', 'n', 1)]}]}
    """
    
    def __init__(self, pos, senses):
        """
        L.__init__(P, S)

        Create a new lemma with part-of-speech P and senses S.
        """
        dict.__init__(self)
        self[pos] =  [{} for i in range(senses)]

    def extend_senses(self, pos, sense):
        """
        L.extend_senses(P, S)

        If the number of senses for words with part-of-speech P in
        lemma L is smaller than S, extend it to include sense S.
        If P does not exist, it will be created.
        """
        if not pos in self:
            self[pos] =  [{} for i in range(sense)]
        elif sense > len(self[pos]):
            self[pos] += [{} for i in range(sense - len(self[pos]))]

    def add_relation(self, pos, sense, name, dest_word, dest_pos, dest_sense):
        """
        L.add_relation(P, S, R, RW, RP, RS)

        Add relation R to word with part-of-speech P and sense S in
        lemma S, where the related word RW has part-of-speech RP and
        sense RS. It is not checked if the relation already exists. 
        """
        try:
            # if not (dest_word, dest_sense) in self[sense - 1][relation]:
            self[pos][sense - 1][name].append((dest_word, dest_pos, dest_sense))
        except KeyError:
            self[pos][sense - 1][name] = [(dest_word, dest_pos, dest_sense)]

    def has_relation(self, relation, other_word, partsos='', senses='',
                   other_pos='', other_sense=''):
        """
        L.has_relation(R, W, [PS='', SS='', P='', S='']) -> True or False
        
        Test if lemma L has relation R to other word W, where L is
        optionally restricted by multiple parts of speech PS and/or
        senses SS, and W is optionally restricted by a part of speech
        P and/or sense S. PS and SS must be comma-separated strings.
        """
        if partsos:
            partsos = partsos.split(',')
        if senses:
            senses = [int(s) for s in senses.split(',')]
        if other_sense:
            other_sense = int(other_sense)
            
        for pos in self.keys():
            if not partsos or pos in partsos:
                for sen in range(1, len(self[pos]) + 1):
                    if not senses or sen in senses:
                        try:
                            for w, p, s in self[pos][sen - 1][relation]:
                                if ( w == other_word and
                                     (not other_pos or p == other_pos) and
                                     (not other_sense or s == other_sense) ):
                                    return True
                        except KeyError:
                            pass
        return False

    def related(self, relation, partsos='', senses=''):
        """
        L.related(R, [PS='', SS='']) -> list of related (word,pos,sense) tuples

        Retrieve all words that have relation R to lemma L, where L is
        optionally restricted by multiple parts of speech PS and/or
        senses SS. PS and SS must be comma-separated strings. The
        related words are returned as a tuple together with their part
        of speech and sense. The returned list may be empty.        
        """
        if partsos:
            partsos = partsos.split(',')
        if senses:
            senses = [int(s) for s in senses.split(',')]
        related = []
       
        for pos in self.keys():
            if not partsos or pos in partsos:
                for sen in range(1, len(self[pos]) + 1):
                    if not senses or sen in senses:
                        try:
                            related += self[pos][sen - 1][relation]
                        except KeyError:
                            pass
        return related
        
    def write(self, s=stdout, indent=0):
        """
        L.write([S=stdout,I=0])

        Write lemma L to stream S with identation I
        """
        for pos in self:
            s.write(indent * ' ')
            s.write('Part-of-speech: %s\n' % pos)
            for sense in range(1, len(self[pos]) + 1):
                self.write_sense(pos, sense, s, indent + 2)

    def write_sense(self, pos, sense, s=stdout, indent=0):
        """
        L.write(P, S, [ST=stdout,I=0])

        Write all senses in lemma L with part-of-speech P to stream ST
        with identation I
        """
        s.write(indent * ' ')
        s.write('Sense: %d\n' % sense)
        for rel in self[pos][sense -1]:
            self.write_relation(pos, sense, rel, s, indent + 2)

    def write_relation(self, pos, sense, name, s=stdout, indent=0):
        """
        L.write(P, S, R, [ST=stdout,I=0])

        Write relation R in lemma L with part-of-speech P and sense S
        to stream ST with identation I
        """
        s.write(indent * ' ')
        s.write('Relation: %s:\n' % name)
        for tuple in self[pos][sense - 1][name]:
            s.write((indent + 2) * ' ')
            s.write('%s (%s,%d)\n' % tuple)


        
def pickle_ewn_files(infns, outfn):
    """
    pickle_ewn_files(INFNS, OUTFN)
    
    Reads the files in Eurowordnet (.ewn) format from list INFNS and
    writes them as single pickled dictionary to file OUTFN (which can
    be reloaded using the load() method from the Lexicon class).
    """
    lex = Lexicon()
    for fn in infns:
        stderr.write('Reading %s\n' % fn)
        lex.read_ewn_file(fn)
    stderr.write('Writing %s\n' % outfn)
    lex.dump(outfn)
    
