# ewnparse.py
# parser for Eurowordnet database files in text format
#
# Copyright (C) 2004  Erwin Marsi
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
# Informationn on how to contact the author can be found in the file CONTACT.

"""
Parser for Eurowordnet files

A library to parse Eurowordnet database files in text format (.ewn)
and transduce them to a list of WORD_MEANING objects
(a.k.a. 'synsets'), so you can do something useful with it. The
structure of a WORD_MEANING object consists of a number of embedded
dicts and lists, roughly equivalent to the indented section structure
in the input file.

Here's an example. Suppose we have a file called 'try.ewn'
with a single entry (double quotes are not escaped in the original):

0 @9136@ WORD_MEANING
  1 PART_OF_SPEECH \"a\"
  1 VARIANTS
    2 LITERAL \"Nederlands\"
      3 SENSE 2
      3 STATUS 182956
      3 EXTERNAL_INFO
        4 CORPUS_ID 1
          5 FREQUENCY 8362
        4 SOURCE_ID 2001
    2 LITERAL \"Ned.\"
      3 SENSE 1
      3 STATUS 182920
      3 EXTERNAL_INFO
        4 CORPUS_ID 1
          5 FREQUENCY 0
        4 SOURCE_ID 2001
  1 INTERNAL_LINKS
    2 RELATION \"is_caused_by\"
      3 TARGET_CONCEPT
        4 PART_OF_SPEECH \"v\"
        4 LITERAL \"vernederlandsen\"
          5 SENSE 1
      3 FEATURES
        4 REVERSED
      3 SOURCE_ID 101

We can parse this in the following way:

>>> from ewnparse import parse_ewn_file
>>> wml = parse_ewn_file('try.ewn')
>>> print wml
[{'VARIANTS': [{'STATUS': 182956, 'EXTERNAL_INFO': {'SOURCE_ID': 2001,
'CORPUS_ID': {'FREQUENCY': 8362, 'ID': 1}}, 'STRING': 'Nederlands',
'SENSE': 2}, {'STATUS': 182920, 'EXTERNAL_INFO': {'SOURCE_ID': 2001,
'CORPUS_ID': {'FREQUENCY': 0, 'ID': 1}}, 'STRING': 'Ned.', 'SENSE':
1}], 'PART_OF_SPEECH': 'a', 'ID': 9136, 'INTERNAL_LINKS':
[{'TARGET_CONCEPT': {'LITERAL': {'STRING': 'vernederlandsen', 'SENSE':
1}, 'PART_OF_SPEECH': 'v'}, 'SOURCE_ID': 101, 'NAME': 'is_caused_by',
'FEATURES': ['REVERSED']}]}]

To get to the string of the first literal in the synset:

>>> wml[0]['VARIANTS'][0]['STRING']
'Nederlands'

Notes:
- works for Dutch Eurowordnet files; not tested for other languages.
- EQ_LINKS section is ignored now, but should also be parsed 
"""

__author__ = 'Erwin Marsi <e.c.marsi@uvt.nl>'
__version__ = '$Id: ewnparse.py,v 1.2 2004/07/13 19:26:22 emarsi Exp $'


# TODO:
# - EQ_LINKS section is ignored now, but should also be parsed 


# In nw_nl_other.ewn, there are a number of variants with *negative* senses
# (rond, vierkant, plat, dun, dik, lang, zoet, zuur, bitter).
# These are probably mistakes (?). If SANE_SENSE is true,
# negative values will be corrected to positive values during parsing.

SANE_SENSE = True


# some trivial classes corresponding to sections in the .ewn file 

class WORD_MEANING(dict):
    pass

class LITERAL(dict):
    pass

class CORPUS_ID(dict):
    pass

class EXTERNAL_INFO(dict):
    pass

class RELATION(dict):
    pass

class TARGET_CONCEPT(dict):
    pass



class SmartFile(file):
    """
    A file object for reading a file which stores the current line,
    and adds a few dummy blank lines at the end of the file.
    """
    
    def __init__(self, fn):
        """
        x.__init__(...) initializes x
        """
        file.__init__(self, fn)
        self.cl = None
        self.blanks = 0

    def next(self):
        """
        x.next() -> the next line, or raise StopIteration

        Reads and strips the next line, and stores the current line in x.cl
        After the real eof is found, it will continue to return 3 blank lines
        until a StopIteration exception is raised.
        """
        try:
            # store current line
            self.cl = file.next(self).strip()
            #print self.cl
            return self.cl
        except StopIteration:
            # eof found, but we're going to return a few blank lines
            # to enable the parsing procedures to finish the last entry properly 
            if self.blanks < 3:
                self.blanks += 1
                self.cl = ''
                return self.cl
            else:
                raise

def parse_ewn_file(fn):
    """
    Parses a file in Eurowordnet (ewn) format and returns it as
    a list of WORD_MEANING dicts.
    """
    f = SmartFile(fn)
    wml = []
    f.next()
    try:
        while True:
            while not f.cl:
                f.next()
            wml.append(parse_word_meaning(f))
    except StopIteration:
        f.close()
        return wml
        

def parse_word_meaning(f):
    """
    Parses a WORD_MEANING section and returns it as
    a WORD_MEANING dict.
    """
    word_meaning = WORD_MEANING()

    # the id of the senseset is stored under key 'ID'
    id = int(f.cl.split()[1].strip('@'))
    word_meaning['ID'] = id
    f.next()

    word_meaning['PART_OF_SPEECH'] = f.cl.split('"')[1]
    f.next()

    if f.cl == '1 VARIANTS':
        f.next()
        word_meaning['VARIANTS'] = []
        while f.cl.startswith('2 LITERAL'):
             word_meaning['VARIANTS'].append(parse_literal(f))

    if f.cl == '1 INTERNAL_LINKS':
        f.next()
        word_meaning['INTERNAL_LINKS'] = []
        while f.cl.startswith('2 RELATION'):
             word_meaning['INTERNAL_LINKS'].append(parse_relation(f))


    if f.cl == '1 EQ_LINKS':
        f.next()
        # word_meaning['EQ_LINKS']  = parse_eq_links(f)
        parse_eq_links(f)
        
    return word_meaning


def parse_literal(f):
    """
    Parses a LITERAL section and returns it as
    a LITERAL dict.
    """
    literal = LITERAL()

    # the literal itself is stored under key 'STRING'
    literal['STRING'] = f.cl.split('"')[1]
    f.next()

    if f.cl.startswith('3 SENSE'):
        sense =  int(f.cl.split()[2])
        if SANE_SENSE:
            # change negative to positive senses
            literal['SENSE'] = abs(sense)
        else:
            literal['SENSE'] = sense
            
        f.next()
        
    if f.cl.startswith('3 DEFINITION'):
        literal['DEFINITION'] = f.cl.split('"')[1]
        f.next()
        
    if f.cl.startswith('3 STATUS'):
        literal['STATUS'] = int(f.cl.split()[2])
        f.next()

    if f.cl.startswith('3 FEATURES'):
        literal['FEATURES'] = []
        f.next()
        while f.cl.startswith('4 FEATURE'):
            feat =  f.cl.split('"')[1]
            f.next()
            # feature values are rare, so we don't allocate another dict,
            # but rather append the value to its feature
            if f.cl.startswith('5 FEATURE_VALUE'):
                feat +=  '=' + f.cl.split('"')[1]
                f.next()
            literal['FEATURES'].append(feat)
            
        
    if f.cl == '3 EXTERNAL_INFO':
        literal['EXTERNAL_INFO'] = EXTERNAL_INFO()
        f.next()

        if f.cl.startswith('4 CORPUS_ID'):
            literal['EXTERNAL_INFO']['CORPUS_ID'] = CORPUS_ID()
            literal['EXTERNAL_INFO']['CORPUS_ID']['ID'] =  int(f.cl.split()[2])
            f.next()

            if f.cl.startswith('5 FREQUENCY'):
                literal['EXTERNAL_INFO']['CORPUS_ID']['FREQUENCY'] =  int(f.cl.split()[2])
                f.next()

        if f.cl.startswith('4 SOURCE_ID'):
            literal['EXTERNAL_INFO']['SOURCE_ID'] = int(f.cl.split()[2])
            f.next()
            
    return literal
    
def parse_relation(f):
    """
    Parses a RELATION section and returns it as
    a RELATION dict.
    """
    relation = RELATION()

    # the name of the relation is stored under key 'NAME' 
    relation['NAME'] = f.cl.split('"')[1]
    f.next()

    if f.cl == '3 TARGET_CONCEPT': 
        relation['TARGET_CONCEPT'] = TARGET_CONCEPT()
        f.next()

        if f.cl.startswith('4 PART_OF_SPEECH'):
            relation['TARGET_CONCEPT']['PART_OF_SPEECH'] = f.cl.split()[-1].strip('"')
            f.next()

        # cannot reuse parse_literal here, because it will assume that '3 FEATURES'
        # belongs to LITERAL, whereas it is in fact part of RELATION
        if f.cl.startswith('4 LITERAL'):
            relation['TARGET_CONCEPT']['LITERAL'] = LITERAL()
            # the literal itself is stored under key 'STRING'
            relation['TARGET_CONCEPT']['LITERAL']['STRING'] = f.cl.split('"')[1]
            f.next()
            # assume that there is always & only a sense with the literal in a relation
            relation['TARGET_CONCEPT']['LITERAL']['SENSE'] = int(f.cl.split()[2])
            f.next()

    if f.cl.startswith('3 FEATURES'):
        relation['FEATURES'] = []
        f.next()
        while f.cl.startswith('4'):
            feat =  f.cl.split()[1]
            relation['FEATURES'].append(feat)
            f.next()
            
    if f.cl.startswith('3 SOURCE_ID'):
        relation['SOURCE_ID'] = int(f.cl.split()[2])
        f.next()

    return relation
    

def parse_eq_links(f):
    eq_links = []
    while f.cl and int(f.cl[0]) > 1:
        f.next()
    return eq_links



